//=============================================================================
//
// Copyright (c) 2009, Sony Pictures Imageworks
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the
// distribution.  Neither the name of Sony Pictures Imageworks nor the
// names of its contributors may be used to endorse or promote
// products derived from this software without specific prior written
// permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
//=============================================================================
#include <maya/MFnNumericAttribute.h>
#include <maya/MFnEnumAttribute.h>
#include <maya/MFnCompoundAttribute.h>
#include <maya/MFnUnitAttribute.h>
#include <maya/MFnTypedAttribute.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MDagPath.h>
#include <maya/MFnCamera.h>
#include <maya/MMatrix.h>
#include <maya/MFnMatrixData.h>
#include <maya/MFnStringData.h>
#include <maya/MTime.h>
#include <maya/MGlobal.h>
#include <maya/MFileIO.h>
#include <maya/MFileObject.h>
#include <maya/MAnimControl.h>
#include <algorithm>
#include <time.h>
#include <gl/glut.h>
#include "reticle.h"


MTypeId Reticle::id(0x00121EC8);
MString Reticle::name("reticle");
MObject Reticle::aDrawingEnabled;
MObject Reticle::aEnableTextDrawing;
MObject Reticle::aFilmbackAperture;
MObject Reticle::aHorizontalFilmAperture;
MObject Reticle::aVerticalFilmAperture;
MObject Reticle::aAspectRatios;
MObject Reticle::aAspectRatio;
MObject Reticle::aDisplayMode;
MObject Reticle::aAspectMaskColor;
MObject Reticle::aAspectMaskTrans;
MObject Reticle::aAspectLineColor;
MObject Reticle::aAspectLineTrans;
MObject Reticle::aAspectDisplaySafeAction;
MObject Reticle::aAspectDisplaySafeTitle;
MObject Reticle::aDisplayLineH;
MObject Reticle::aDisplayLineV;
MObject Reticle::aDisplayThirdsH;
MObject Reticle::aDisplayThirdsV;
MObject Reticle::aDisplayCrosshair;
MObject Reticle::aLineColor;
MObject Reticle::aLineTrans;
MObject Reticle::aTime;
MObject Reticle::aText;
MObject Reticle::aTextType;
MObject Reticle::aTextStr;
MObject Reticle::aTextSize;
MObject Reticle::aTextAlign;
MObject Reticle::aTextPos;
MObject Reticle::aTextPosX;
MObject Reticle::aTextPosY;
MObject Reticle::aTextPosRel;
MObject Reticle::aTextLevel;
MObject Reticle::aTextARLevel;
MObject Reticle::aTextColor;
MObject Reticle::aTextTrans;


Reticle::Reticle()
{
}


Reticle::~Reticle()
{
}


MPoint Reticle::getPoint(float x, float y, M3dView & view)
{
	// This function takes in a screen x and y value and converts it to values
	// on the near and far clipping planes.
	MPoint p, ncp, fcp;

	view.viewToWorld(int(x), int(y), ncp, fcp);
	MVector v = fcp - ncp;
	p = (ncp + v) * wim;

	return p;
}


MStatus Reticle::getColor(MObject colorObj, MObject transObj, MColor &color)
{
	// This method will retrieve the individual r,g,b and alpha values from
	// various plugs and store them in an MColor object.
	MPlug p,c;

	p = MPlug(thisNode, colorObj);

	c = p.child(0);
	c.getValue(color.r);

	c = p.child(1);
	c.getValue(color.g);

	c = p.child(2);
	c.getValue(color.b);

	p = MPlug(thisNode, transObj);
	p.getValue(color.a);

	return MS::kSuccess;
}


MMatrix Reticle::getMatrix(MString matrixStr)
{
	// This method returns the specified matrix.
	MStatus stat;
	MMatrix matrix;

	// Extract the matrix attribute that is inherited from 'dagNode'
	MFnDependencyNode fnThisNode(thisNode);
	MObject matrixAttribute = fnThisNode.attribute(matrixStr, &stat);
	if (!stat) {
		stat.perror("Reticle::getMatrix getting attribute : "+matrixStr);
		return matrix.identity;
	}

	MPlug matrixPlug(thisNode, matrixAttribute);
	if (matrixPlug.isNull()) {
		stat.perror("Reticle::getting plug for attribute : "+matrixStr);
		return matrix.identity;
	}

	// If matrix is an array so we must specify
	// which element the plug refers to
	if (matrixPlug.isArray()) {
		matrixPlug = matrixPlug.elementByLogicalIndex (0, &stat);
		if(!stat)
		{
			stat.perror(
				"Reticle::getMatrix setting logical index to 0 for "+matrixStr
				);
			return matrix.identity;
		}
	}

	// Get the value of the matrix attribute
	MObject matObject;
	stat = matrixPlug.getValue(matObject);
	if (!stat) {
		stat.perror("Reticle::getMatrix getting value for "+matrixStr);
		return matrix.identity;
	}

	MFnMatrixData matrixData(matObject, &stat);
	if (!stat) {
		stat.perror("Reticle::getMatrix getting MfnMatrixData");
		return matrix.identity;
	}

	matrix = matrixData.matrix(&stat);
	if (!stat) {
		stat.perror("Reticle::getMatrix getting matrix");
		return matrix.identity;
	}

	return matrix;
}


MStatus Reticle::getFilmbackData()
{
	// This method retrieves all of the data related to the filmback from
	// various plugs.
	MPlug p;

	// Get horizontal film aperture
	p = MPlug(thisNode, aHorizontalFilmAperture);
	p.getValue(filmback.horizontalFilmAperture);

	// Get vertical film aperture
	p = MPlug(thisNode, aVerticalFilmAperture);
	p.getValue(filmback.verticalFilmAperture);

	return MS::kSuccess;
}


MStatus Reticle::getAspectRatioChildren(MPlug arPlug, Aspect_Ratio &ar)
{
	// This method gets the data for a specified aspect ratio. The aspectratios
	// attribute is effectively a complex compound attribute where each index
	// is the data for a specific aspect ratio.
	MPlug p, c;

	p = arPlug.child(0);
	p.getValue(ar.aspectRatio);

	p = arPlug.child(1);
	p.getValue(ar.displayMode);

	if (ar.displayMode != 0) {
		p = arPlug.child(2);
		c = p.child(0);
		c.getValue(ar.aspectGeom.maskColor.r);

		c = p.child(1);
		c.getValue(ar.aspectGeom.maskColor.g);

		c = p.child(2);
		c.getValue(ar.aspectGeom.maskColor.b);

		p = arPlug.child(3);
		p.getValue(ar.aspectGeom.maskColor.a);

		p = arPlug.child(4);
		c = p.child(0);
		c.getValue(ar.aspectGeom.lineColor.r);

		c = p.child(1);
		c.getValue(ar.aspectGeom.lineColor.g);

		c = p.child(2);
		c.getValue(ar.aspectGeom.lineColor.b);

		p = arPlug.child(5);
		p.getValue(ar.aspectGeom.lineColor.a);

		p = arPlug.child(6);
		p.getValue(ar.displaySafeAction);

		p = arPlug.child(7);
		p.getValue(ar.displaySafeTitle);
	}

	return MS::kSuccess;
}


bool Reticle::aspectRatioSortPredicate(
	const Aspect_Ratio &lhs,
	const Aspect_Ratio &rhs
	)
{
	return lhs.aspectRatio <= rhs.aspectRatio;
}


MStatus Reticle::getAspectRatioData()
{
	// This retrieves the data for all of the various aspect ratios.
	MPlug arsPlug = MPlug(thisNode, aAspectRatios);
	int numPlugs = arsPlug.numElements();

	MStatus stat;

	// Clear the aspect ratio vector
	ars.clear();

	for (int i = 0; i < numPlugs; i++)
	{
		MPlug p = arsPlug.elementByPhysicalIndex(i);

		Aspect_Ratio ar;
		getAspectRatioChildren(p, ar);
		ars.push_back(ar);
	}

	numAspectRatios = (int)ars.size();

	// Sort aspect ratios
	std::sort(ars.begin(), ars.end(), aspectRatioSortPredicate);

	return MS::kSuccess;
}


bool Reticle::needToUpdateAspectRatios()
{
	MPlug arsPlug = MPlug(thisNode, aAspectRatios);
	return (ars.size() != arsPlug.numElements());
}


MStatus Reticle::getTextChildren(MPlug tPlug, TextData &td)
{
	// This method gets the data for a specified text entry. The text attribute
	// is a complex compound attribute where each index is the data for a specific
	// text element.
	MPlug p, c;

	p = tPlug.child(0);
	p.getValue(td.textType);

	p = tPlug.child(1);
	p.getValue(td.textStr);

	p = tPlug.child(2);
	p.getValue(td.textSize);

	p = tPlug.child(3);
	p.getValue(td.textAlign);

	p = tPlug.child(4);
	c = p.child(0);
	 c.getValue(td.textPosX);

	c = p.child(1);
	 c.getValue(td.textPosY);

	p = tPlug.child(5);
	p.getValue(td.textPosRel);

	p = tPlug.child(6);
	 p.getValue(td.textLevel);

	p = tPlug.child(7);
	 p.getValue(td.textARLevel);

	p = tPlug.child(8);
	c = p.child(0);
	c.getValue(td.textColor.r);

	c = p.child(1);
	c.getValue(td.textColor.g);

	c = p.child(2);
	c.getValue(td.textColor.b);

	p = tPlug.child(9);
	p.getValue(td.textColor.a);

	return MS::kSuccess;
}


MStatus Reticle::getTextData()
{
	// Clear the text vector
	text.clear();

	MPlug p,tp,cp;

	p = MPlug(thisNode, aText);
	int numPlugs = p.numElements();

	for (int i = 0; i < numPlugs && i < 10; i++) {
		TextData td;
		tp = p.elementByPhysicalIndex(i);
		getTextChildren(tp, td);
		text.push_back(td);
	}

	return MS::kSuccess;
}


bool Reticle::needToUpdateTextData()
{
	MPlug p = MPlug(thisNode, aText);
	return (text.size() != p.numElements());
}


MStatus Reticle::getOptions()
{
	// This method retrieves all of the options settings.
	MPlug p;

	// Check to see if drawing is enabled
	p = MPlug(thisNode, aDrawingEnabled);
	p.getValue(options.drawingEnabled);

	if (options.drawingEnabled) {
		// Check to see if the text should be rendered;
		p = MPlug(thisNode, aEnableTextDrawing);
		p.getValue(options.enableTextDrawing);

		// Display horizontal line option;
		p = MPlug(thisNode, aDisplayLineH);
		p.getValue(options.displayLineH);

		// Display vertical line option;
		p = MPlug(thisNode, aDisplayLineV);
		p.getValue(options.displayLineV);

		// Display horizontal thirds option;
		p = MPlug(thisNode, aDisplayThirdsH);
		p.getValue(options.displayThirdsH);

		// Display vertical thirds option;
		p = MPlug(thisNode, aDisplayThirdsV);
		p.getValue(options.displayThirdsV);

		// Display crosshair option;
		p = MPlug(thisNode, aDisplayCrosshair);
		p.getValue(options.displayCrosshair);

		// Line Color;
		getColor(aLineColor, aLineTrans, options.lineColor);
	}

	return MS::kSuccess;
}


void Reticle::calcPortGeom()
{
	// This method calculates the geometry of the current window. This data is
	// stored in a instance of the Geom class.
	portGeom.x1 = 0;
	portGeom.x2 = portWidth;
	portGeom.y1 = 0;
	portGeom.y2 = portHeight;

	portGeom.x = portWidth / 2;
	portGeom.y = portHeight / 2;
	portGeom.isValid = true;
}


void Reticle::calcFilmbackGeom()
{
	// This method calculates the geometry for the filmback. There are multiple
	// Geom instances that are used for displaying all of the filmback options.
	// This method takes the filmback data and calculates the appropriate settings
	// for the filmback Geom instances.
	filmback = oFilmback;

	// See if we are using the camera's filmback data
	bool useCameraFilmback = filmback.horizontalFilmAperture == -1;

	// If the horizontalFilmAperture equals -1, then use the cameras settings
	if (useCameraFilmback) {
		filmback.horizontalFilmAperture = camera.horizontalFilmAperture();
		filmback.verticalFilmAperture = camera.verticalFilmAperture();
	}

	double hfa = filmback.horizontalFilmAperture;
	double vfa = filmback.verticalFilmAperture;

	double aspectRatio = hfa/vfa;
	double nonPaddedAspectRatio = aspectRatio;

	double portAspectRatio = double(portWidth)/double(portHeight);
	bool portHoriz = portAspectRatio > aspectRatio;

	MFnCamera::FilmFit filmFit = camera.filmFit();
	if (filmFit == MFnCamera::kFillFilmFit) {
		if (portHoriz)
			filmFit = MFnCamera::kHorizontalFilmFit;
		else
			filmFit = MFnCamera::kVerticalFilmFit;
	}
	if (filmFit == MFnCamera::kOverscanFilmFit) {
		if (portHoriz)
			filmFit = MFnCamera::kVerticalFilmFit;
		else
			filmFit = MFnCamera::kHorizontalFilmFit;
	}

	switch (filmFit) {
		case MFnCamera::kInvalid:
		case MFnCamera::kHorizontalFilmFit:
			filmback.filmbackGeom.x = (portWidth / overscan);
			filmback.filmbackGeom.y = filmback.filmbackGeom.x / nonPaddedAspectRatio;
			break;
		case MFnCamera::kVerticalFilmFit:
			filmback.filmbackGeom.y = (portHeight / overscan);
			filmback.filmbackGeom.x = filmback.filmbackGeom.y * nonPaddedAspectRatio;
			break;
		default:
			break;
	}

	// Account for lens squeeze
	filmback.filmbackGeom.x *= camera.lensSqueezeRatio();

	filmback.filmbackGeom.x1 = portGeom.x - (filmback.filmbackGeom.x / 2);
	filmback.filmbackGeom.x2 = portGeom.x + (filmback.filmbackGeom.x / 2);
	filmback.filmbackGeom.y1 = portGeom.y - (filmback.filmbackGeom.y / 2);
	filmback.filmbackGeom.y2 = portGeom.y + (filmback.filmbackGeom.y / 2);
	filmback.filmbackGeom.isValid = true;

	// Calculate the image area
	filmback.imageGeom = filmback.filmbackGeom;
}


void Reticle::calcMaskGeom(
	Geom &g,
	double w,
	double h,
	Geom &gSrc,
	double wSrc,
	double hSrc
	)
{
	// This method calculates the actual mask x,y values for a given Geom
	// instance.
	double pw = (w != -1) ? ((wSrc-w) / 2) / wSrc : (1 - wSrc) / 2;
	double ph = (h != -1) ? ((hSrc-h) / 2) / hSrc : (1 - hSrc) / 2;

	g.x = gSrc.x * pw;
	g.y = gSrc.y * ph;

	g.x1 = gSrc.x1 + g.x;
	g.x2 = gSrc.x2 - g.x;
	g.y1 = gSrc.y1 + g.y;
	g.y2 = gSrc.y2 - g.y;
}


void Reticle::calcSafeActionGeom(Aspect_Ratio &ar)
{
	// Calculates the safe action geometry for an Aspect Ratio.
	if (ar.aspectGeom.isValid) {
		ar.safeActionGeom = ar.aspectGeom;
		calcMaskGeom(ar.safeActionGeom, -1, -1, ar.aspectGeom, 0.9, 0.9);
		ar.safeActionGeom.isValid = true;

	} else {
		ar.safeActionGeom.isValid = false;
	}
}


void Reticle::calcSafeTitleGeom(Aspect_Ratio &ar)
{
	// Calculates the safe title geometry for an Aspect Ratio.
	if(ar.aspectGeom.isValid) {
		ar.safeTitleGeom = ar.aspectGeom;
		calcMaskGeom(ar.safeTitleGeom, -1, -1, ar.aspectGeom, 0.8, 0.8);
		ar.safeTitleGeom.isValid = true;

	} else {
		ar.safeTitleGeom.isValid = false;
	}
}


void Reticle::calcAspectGeom(Aspect_Ratio &ar)
{
	// This calculates the various Geom instances for a particular aspectRatio
	// instance.
	ar.aspectGeom.x = filmback.imageGeom.x;
	ar.aspectGeom.y = ar.aspectGeom.x / ar.aspectRatio;

	ar.aspectGeom.x1 = filmback.imageGeom.x1;
	ar.aspectGeom.x2 = filmback.imageGeom.x2;
	ar.aspectGeom.y1 = portGeom.y - (ar.aspectGeom.y / 2);
	ar.aspectGeom.y2 = portGeom.y + (ar.aspectGeom.y / 2);

	ar.aspectGeom.isValid = true;
	ar.safeActionGeom.isValid = false;
	ar.safeTitleGeom.isValid = false;

	if (ar.displaySafeAction)
		calcSafeActionGeom(ar);

	if (ar.displaySafeTitle)
		calcSafeTitleGeom(ar);
}


bool Reticle::setInternalValueInContext(
	const MPlug &plug,
	const MDataHandle &dataHandle,
	MDGContext &ctx
	)
{
	if(plug == worldInverseMatrix || plug == isTemplated)
		return false;

	needRefresh = true;

	return false;
}


void Reticle::drawMask(Geom g1, Geom g2, MColor color, bool sides)
{
	// Given two Geom instances,
	// this calculates the mask area between them.
	double z = -ncp;

	// Turn off z-depth test
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	glBegin(GL_QUADS);

		glColor4f(color.r, color.g, color.b, 1-color.a);

		// Bottom Mask
		glVertex3d(g1.x1, g1.y1, z);
		glVertex3d(g1.x2, g1.y1, z);
		glVertex3d(g2.x2, g2.y1, z);
		glVertex3d(g2.x1, g2.y1, z);

		// Top Mask
		glVertex3d(g2.x1, g2.y2, z);
		glVertex3d(g2.x2, g2.y2, z);
		glVertex3d(g1.x2, g1.y2, z);
		glVertex3d(g1.x1, g1.y2, z);

		if (sides) {
			// Left side mask
			glVertex3d(g1.x1, g1.y1, z);
			glVertex3d(g2.x1, g2.y1, z);
			glVertex3d(g2.x1, g2.y2, z);
			glVertex3d(g1.x1, g1.y2, z);

			// right side mask
			glVertex3d(g2.x2, g2.y1, z);
			glVertex3d(g1.x2, g1.y1, z);
			glVertex3d(g1.x2, g1.y2, z);
			glVertex3d(g2.x2, g2.y2, z);
		}

	glEnd();

	// Turn on z-depth test
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}


void Reticle::drawLine(
	double x1, double x2,
	double y1, double y2,
	MColor color,
	bool stipple
	)
{
	// This draws a single line between the specified points.
	double z = -ncp;

	// Turn off z-depth test
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	if (stipple) {
		glEnable(GL_LINE_STIPPLE);
		glLineStipple(2, 0x00FF);
	}

	glBegin(GL_LINES);

		glColor4f(color.r, color.g, color.b, 1-color.a);

		glVertex3d(x1, y1, z);
		glVertex3d(x2, y2, z);

	glEnd();

	if (stipple)
		glDisable(GL_LINE_STIPPLE);

	// Turn on z-depth test
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}


void Reticle::drawLines(Geom g, MColor color, bool sides, bool stipple)
{
	// Given a Geom instance, this will draw a line connecting the points.
	// The argument side determines whether the sides will be drawn (the top
	// will always be drawn). The stipple argument specifies whether the line
	// should be solid or dashed/stippled.
	double z = -ncp;

	//Turn off z-depth test
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	if (stipple) {
		glEnable(GL_LINE_STIPPLE);
		glLineStipple(2, 0x00FF);
	}

	GLenum mode = (sides) ? GL_LINE_LOOP : GL_LINES;

	glBegin(mode);

		glColor4f(color.r, color.g, color.b, 1-color.a);

		glVertex3d(g.x1, g.y1, z);
		glVertex3d(g.x2, g.y1, z);
		glVertex3d(g.x2, g.y2, z);
		glVertex3d(g.x1, g.y2, z);

	glEnd();

	if (stipple)
		glDisable (GL_LINE_STIPPLE);

	//Turn on z-depth test
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}


void Reticle::drawText(
	MString text,
	double tx, double ty,
	void *font, MColor textColor, int textAlign, int textValign,
	M3dView & view
	)
{
	// This function uses Maya's built-in function to draw text. At somepoint
	// Turn off z-depth test
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	// Custom
	const char *strA = text.asChar();
	const char *strB = text.asChar();

	// Get width
	int width = 0;
	while (*strA) {
		width += glutBitmapWidth(font, *strA);
		++strA;
	}

	// Get height
	int height = 10;
	if (font == GLUT_BITMAP_HELVETICA_10) {
		height = 10;
	
	} else if (font == GLUT_BITMAP_HELVETICA_12) {
		height = 12;
	
	} else if(font == GLUT_BITMAP_HELVETICA_18) {
		height = 18;
	}

	// align
	if(textAlign == 1) {			// left
		tx -= width / 2;
	
	} else if(textAlign == 2) {	// right
		tx -= width;
	}

	// valign
	if (textValign == 0) {		// top
		ty -= height;
	
	} else if(textValign == 1) {	// Middle
		ty -= height / 2;
	
	} else { 						// Bottom
		ty += 3;// ajustment
	}

	MPoint textPos = getPoint((float)tx, (float)ty, view);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glColor4f(textColor.r, textColor.g, textColor.b, 1-textColor.a);
	glPushMatrix();
		glTranslatef(textPos.x, textPos.y, textPos.z);
		glRasterPos3f(0, 0, 0);
		while (*strB) {
			glutBitmapCharacter(font, *strB);
			++strB;
		}
	glPopMatrix();
	glFlush();

	// Turn on z-depth test
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}


void Reticle::drawCustomTextElements(M3dView &view)
{
	TextData td;
	char buff[255];

	double start = MAnimControl::minTime().value();
	double end   = MAnimControl::maxTime().value();


	MTime mtime;
	MPlug p = MPlug(thisNode, aTime);
	if (p.isNull() == false && p.isConnected() == false) {
		MString cmd = "connectAttr time1.o "+p.name();
		MGlobal::executeCommand(cmd);
	}
	p.getValue(mtime);

	for (int i = 0; i<(int)text.size(); i++) {
		td = text[i];
		switch(td.textType) {
			case 0:						//String
				break;
			case 1:						//Lens
			{
				if(td.textStr == "")
					td.textStr = MString("%1.2f mm");

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										camera.focalLength()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");
				
				td.textStr = MString(buff);
				break;
			}
			case 2:						//Camera
			{
				if(td.textStr == "")
					td.textStr = MString("%s");

				MFnDependencyNode cameraParent(camera.parent(0));
				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										cameraParent.name().asChar()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");
				
				td.textStr = MString(buff);
				break;
			}
			case 3:						//Frame
			{
				if(td.textStr == "")
					td.textStr = MString("%04.0f");

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										mtime.value()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 4:						//Frame And Range
			{
				if(td.textStr == "")
					td.textStr = MString("%04.0f (%04.0f - %04.0f)");
				
				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										mtime.value(),
										start,
										end
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 5:						//Frame Start
			{
				if(td.textStr == "")
					td.textStr = MString("%04.0f");

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										start
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 6:						//Frame End
			{
				if(td.textStr == "")
					td.textStr = MString("%04.0f");

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										end
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 7:						//Frame Range
			{
				if(td.textStr == "")
					td.textStr = MString("%04.0f - %04.0f");
				
				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										start,
										end
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 8:						//User
			{
				if(td.textStr == "")
					td.textStr = MString("%s");

				char   *pBuffer;
				size_t pSize;
				errno_t err = _dupenv_s(&pBuffer, &pSize, "USER");
				if(err)
				{
					pBuffer = "";
				}

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										pBuffer
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");

				td.textStr = MString(buff);
				break;
			}
			case 9:						//Current File
			{
				if (td.textStr == "")
					td.textStr = MString("%s");

				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										MFileIO::currentFile().asChar()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");
				
				td.textStr = MString(buff);
				break;
			}
			case 10:						//Path
			{
				if (td.textStr == "")
					td.textStr = MString("%s");

				MFileObject fileObject;
				fileObject.setFullName(MFileIO::currentFile());
				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										fileObject.path().asChar()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");
				
				td.textStr = MString(buff);
				break;
			}
			case 11:						//Path
			{
				if (td.textStr == "")
					td.textStr = MString("%s");

				MFileObject fileObject;
				fileObject.setFullName(MFileIO::currentFile());
				int result = sprintf_s(	buff,
										td.textStr.asChar(),
										fileObject.name().asChar()
										);
				if(!result)
					strcpy_s(buff, sizeof(buff), "Error");
				
				td.textStr = MString(buff);
				break;
			}
			case 12:						//Date
			{
				time_t timer = time(NULL);
				struct tm date;
				localtime_s(&date, &timer);
				const char *format = "%b, %d, %Y(%a)  %I:%M %p";

				if(td.textStr != "")
					format = td.textStr.asChar();

				strftime(buff, 255, format, &date);

				td.textStr = MString(buff);

				break;
			}
			default:
				continue;
		}

		// Determine the level geometry
		Geom g;
		switch (td.textLevel) {
			case 0:
				g = portGeom;
				break;
				
			case 1:
				g = filmback.filmbackGeom;
				break;
				
			case 2:
			{
				int level = td.textARLevel;
				if (level < 0 || level >= numAspectRatios)
					continue;

				if (!ars[level].aspectGeom.isValid)
					continue;

				g = ars[level].aspectGeom;
				break;
			}
			default:
				break;
		}

		// Determine the position
		double x = 0;
		double y = 0;

		switch (td.textPosRel) {
			case 0:						//Bottom Left
				x = g.x1;
				y = g.y1;
				break;
				
			case 1:						//Bottom Center
				x = (g.x1 + g.x2) / 2;
				y = g.y1;
				break;
				
			case 2:						//Bottom Right
				x = g.x2;
				y = g.y1;
				break;
				
			case 3:						//Middle Left
				x = g.x1;
				y = (g.y1 + g.y2) / 2;
				break;
				
			case 4:						//Center
				x = (g.x1 + g.x2) / 2;
				y = (g.y1 + g.y2) / 2;
				break;
				
			case 5:						//Middle Right
				x = g.x2;
				y = (g.y1 + g.y2) / 2;
				break;
				
			case 6:						//Top Left
				x = g.x1;
				y = g.y2;
				break;
				
			case 7:						//Top Center
				x = (g.x1 + g.x2) / 2;
				y = g.y2;
				break;
				
			case 8:						//Top Right
				x = g.x2;
				y = g.y2;
				break;
				
			default:
				continue;
		}

		// Drawing without flicker.
		if (x <= 0)
			x = 1;

		// Valign
		int textValign = 2;
		if (td.textLevel == 0 || td.textLevel == 1) {
		
			if (td.textPosRel >= 6 && td.textPosRel <= 8)
				textValign = 0;
		
		} else if (td.textLevel == 2) {
			if (td.textPosRel >= 0 && td.textPosRel <= 2)
				textValign = 0;
		}


		// Add text position as offset
		x += td.textPosX;
		y += td.textPosY;
		
		// Font
		void *font;
		if(td.textSize == 0)
			font = GLUT_BITMAP_HELVETICA_10;

		else if(td.textSize == 1)
			font = GLUT_BITMAP_HELVETICA_12;

		else if(td.textSize == 2)
			font = GLUT_BITMAP_HELVETICA_18;


		// Actually draw the text
		drawText(	td.textStr,
					x, y,
					font, td.textColor, td.textAlign, textValign,
					view
					);
	}
}


void Reticle::draw(
	M3dView &view,
	const MDagPath &path,
	M3dView::DisplayStyle style,
	M3dView::DisplayStatus status
	)
{
	// This is the main function which draws the locator. It is called
	// by Maya whenever a 3D view needs to be refreshed.
	MStatus stat;
	MPlug p;
	thisNode = thisMObject();

	// Get options
	if(needRefresh)
		getOptions();

	// Drawing not enabled, return
	if(!options.drawingEnabled)
		return;

	// Get the dagPath,
	// node and function set for the currently displaying camera
	MDagPath cameraPath;
	view.getCamera(cameraPath);

	// Set the MFnCamera to the current camera
	camera.setObject(cameraPath);

	// If camera is orthographic, then return
	if(camera.isOrtho()) 
		return;

	// Using useSpReticle Filter, check camera for attribute
	{
		//Default value to false
		int useReticle = false;

		// Find the useSpReticle plug on the camera node
		MPlug useReticlePlug = camera.findPlug("useReticle", &stat);

		// If the plug is valid (not null) get the current value
		if (!useReticlePlug.isNull())
			useReticlePlug.getValue(useReticle);
		else
			useReticle = true;

		// If camera is not set to use reticle then return
		if (!useReticle)
			return;
	}

	// Get the camera position
	MMatrix wm = cameraPath.inclusiveMatrix();

	// Get the camera's overscan value
	overscan = camera.overscan();
	ncp = camera.nearClippingPlane() + 0.001;

	// Get the view attributes that impact drawing
	portWidth = double (view.portWidth());
	portHeight = double (view.portHeight());

	// Get the worldInverseMatrix
	wim = getMatrix("worldInverseMatrix");

	// Calculate the port geometry
	calcPortGeom();

	// Update text data
	if(needToUpdateTextData() || needRefresh)
		getTextData();

	// Update aspect ratios
	if(needToUpdateAspectRatios() || needRefresh)
		getAspectRatioData();

	// Check to see if we need to refresh the data
	if (needRefresh) {
		getFilmbackData();		// Get the filmback attribute data
		oFilmback = filmback;	// Hold this copy
		needRefresh = false;	// Reset need refresh
	}

	// Peform the calculations necessary to define the filmback geometry.
	calcFilmbackGeom();

	// Calculate the aspect ratio geometry
	for (int i = 0; i < numAspectRatios; i++) {
		// Only calculate the aspect geom if the display mode != 0
		if (ars[i].displayMode != 0)
			calcAspectGeom(ars[i]);
	}

	// Begin drawing locator
	view.beginGL();

	// Store all of the openGL attribute settings
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	// Go into 2D ortho mode
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(
		0.0, (GLdouble)view.portWidth(),
		0.0, (GLdouble)view.portHeight(),
		camera.nearClippingPlane(),
		camera.farClippingPlane()
		);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Turn on openGL blending for transparency
	glEnable(GL_BLEND);

	// Store the current blend types
	GLint blendAttrs[2];
	glGetIntegerv(GL_BLEND_SRC, & blendAttrs[0]);
	glGetIntegerv(GL_BLEND_DST, & blendAttrs[1]);

	// Set the blending function
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// create variable to store what
	// the first geometry object to draw aspect ratios to
	Geom aspectContainerGeom;
	aspectContainerGeom = portGeom;

	// Draw the masks first
	Aspect_Ratio ar;
	for (int i = 0; i < numAspectRatios; i++) {
		ar = ars[i];
		// Draw the masks as Quads
		if (ar.displayMode == 2) {
			Geom g = (i == 0) ? aspectContainerGeom : ars[i-1].aspectGeom;
			drawMask(g, ar.aspectGeom, ar.aspectGeom.maskColor, i == 0);
		}
	}

	// Draw lines
	for (int i = 0; i < numAspectRatios; i++) {
		ar = ars[i];

		if (ar.displayMode != 0) {
			bool sides = (i == 0 || ar.aspectGeom.x != ars[i].aspectGeom.x);
			drawLines(	ar.aspectGeom,
						ar.aspectGeom.lineColor,
						sides,
						0
						);

			// Draw safe action
			if(ar.displaySafeAction)
				drawLines(	ar.safeActionGeom,
							ar.safeActionGeom.lineColor,
							1,
							1
							);

			// Draw safe title
			if(ar.displaySafeTitle)
				drawLines(	ar.safeTitleGeom,
							ar.safeTitleGeom.lineColor,
							1,
							1
							);
		}
	}

	// Display horizontal line
	if(options.displayLineH) {
		drawLine(	filmback.imageGeom.x1,
					filmback.imageGeom.x2,
					portGeom.y,
					portGeom.y,
					options.lineColor,
					0
					);
	}

	// Display vertical line
	if (options.displayLineV) {
		double cx = (filmback.imageGeom.x1 + filmback.imageGeom.x2) / 2;
		drawLine(	cx,
					cx,
					filmback.imageGeom.y1,
					filmback.imageGeom.y2,
					options.lineColor,
					0
					);
	}

	// Display Horizontal Thirds
	if (options.displayThirdsH) {
		Geom g = (numAspectRatios > 0)?ars[0].aspectGeom:portGeom;
		double y1 = g.y1+ ((g.y2 - g.y1) * 0.33);
		double y2 = g.y1+ ((g.y2 - g.y1) * 0.66);
		drawLine(g.x1, g.x2, y1, y1, options.lineColor, 0);
		drawLine(g.x1, g.x2, y2, y2, options.lineColor, 0);
	}

	// Display Vertical Thirds
	if (options.displayThirdsV) {
		Geom g = (numAspectRatios > 0)?ars[0].aspectGeom:portGeom;
		double x1 = g.x1+ ((g.x2 - g.x1) * 0.33);
		double x2 = g.x1+ ((g.x2 - g.x1) * 0.66);
		drawLine(x1, x1, g.y1, g.y2, options.lineColor, 0);
		drawLine(x2, x2, g.y1, g.y2, options.lineColor, 0);
	}

	// Display crosshair
	if (options.displayCrosshair) {
		double cx = (filmback.imageGeom.x1 + filmback.imageGeom.x2) / 2;
		drawLine(cx-25, cx-5, portGeom.y, portGeom.y, options.lineColor, 0);
		drawLine(cx+25, cx+5, portGeom.y, portGeom.y, options.lineColor, 0);
		drawLine(cx, cx, portGeom.y-25, portGeom.y-5, options.lineColor, 0);
		drawLine(cx, cx, portGeom.y+25, portGeom.y+5, options.lineColor, 0);
	}

	// Draw custom text elements
	if(options.enableTextDrawing)
		drawCustomTextElements(view);

	// Turn off blending
	glDisable(GL_BLEND);

	// Restore blend settings
	glBlendFunc(blendAttrs[0], blendAttrs[1]);

	// Restore matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	// Restore all attributes
	glPopAttrib();

	// End of openGL calls
	view.endGL();
}


bool Reticle::isBounded() const
{ 
	return true;
}


bool Reticle::isTransparent() const
{
	return true;
}


bool Reticle::drawLast() const
{
	return true;
}


MBoundingBox Reticle::boundingBox() const
{
	return MBoundingBox(	MPoint(-1000000,-1000000,-1000000),
							MPoint(1000000,1000000,1000000)
							);
}


bool Reticle::excludeAsLocator() const
{
	// This function returns whether this locator should be effected by the
	// Display locators toggle in Maya. It is dependent upon the hideLocator
	// option.
	return false;
}


void Reticle::postConstructor()
{
	// This method is called just after a particular instance of the node has
	// been created. It automatically sources the proper scripts and sets the
	// loadDefault argument to 1, indicating that the first time the locator
	// is drawn that it should first load the default settings for the locator.
	// This functionality allows either the show or facility to set standards
	// on the locator.

	// Tell Maya to draw this in the transparent queue
#if MAYA_API_VERSION < 850
#ifndef WIN32
	setTranspHandler((TranspHandler)(&Reticle::isTransparent));
#endif
#endif

	// Set Refresh
	needRefresh = true;
}


void * Reticle::creator()
{
	return new Reticle();
}


MStatus Reticle::initialize()
{
	MFnNumericAttribute		nAttr;
	MFnEnumAttribute		eAttr;
	MFnCompoundAttribute	cAttr;
	MFnUnitAttribute		uAttr;
	MFnTypedAttribute		tAttr;

	// Define a default empty string
	MString defaultString("");
	MFnStringData defaultTextData;
	MObject defaultTextAttr = defaultTextData.create(defaultString);

	aDrawingEnabled = nAttr.create(
						"drawingEnabled",
						"en",
						MFnNumericData::kBoolean,
						1
						);
	nAttr.setInternal(true);

	aEnableTextDrawing = nAttr.create(
							"enableTextDrawing",
							"etd",
							MFnNumericData::kBoolean,
							1
							);
	nAttr.setInternal(true);

	aHorizontalFilmAperture = nAttr.create(
								"horizontalFilmAperture",
								"hfa",
								MFnNumericData::kFloat,
								1.778
								);
	nAttr.setInternal(true);

	aVerticalFilmAperture = nAttr.create(
								"verticalFilmAperture",
								"vfa",
								MFnNumericData::kFloat,
								1.0
								);
	nAttr.setInternal(true);

	aFilmbackAperture = nAttr.create(
							"filmbackAperture",
							"cap",
							aHorizontalFilmAperture,
							aVerticalFilmAperture
							);

	aAspectRatio = nAttr.create(
							"aspectRatio",
							"ar",
							MFnNumericData::kFloat,
							1.85
							);
	nAttr.setInternal(true);

	aDisplayMode = eAttr.create("displayMode", "dm", 2);
	eAttr.addField("none", 0);
	eAttr.addField("lines", 1);
	eAttr.addField("Transparent Mask", 2);
	eAttr.setInternal(true);

	aAspectMaskColor = nAttr.createColor("aspectMaskColor", "amc");
	nAttr.setDefault(0.0, 0.0, 0.0);
	nAttr.setUsedAsColor(true);
	nAttr.setInternal(true);

	aAspectMaskTrans = nAttr.create(
							"aspectMaskTrans",
							"amt",
							MFnNumericData::kFloat,
							0.75
							);
	nAttr.setMin(0.0);
	nAttr.setMax(1.0);
	nAttr.setInternal(true);

	aAspectLineColor = nAttr.createColor("aspectLineColor", "alc");
	nAttr.setDefault(0.8, 0.8, 0.8);
	nAttr.setUsedAsColor(true);
	nAttr.setInternal(true);

	aAspectLineTrans = nAttr.create(
							"aspectLineTrans",
							"alt",
							MFnNumericData::kFloat,
							0.0
							);
	nAttr.setMin(0.0);
	nAttr.setMax(1.0);
	nAttr.setInternal(true);

	aAspectDisplaySafeTitle = nAttr.create(
								"aspectDisplaySafeTitle",
								"adst",
								MFnNumericData::kBoolean,
								false
								);
	nAttr.setInternal(true);

	aAspectDisplaySafeAction = nAttr.create(
								"aspectDisplaySafeAction",
								"adsa",
								MFnNumericData::kBoolean,
								false
								);
	nAttr.setInternal(true);

	aAspectRatios = cAttr.create("aspectRatios", "ars");
	cAttr.addChild(aAspectRatio);
	cAttr.addChild(aDisplayMode);
	cAttr.addChild(aAspectMaskColor);
	cAttr.addChild(aAspectMaskTrans);
	cAttr.addChild(aAspectLineColor);
	cAttr.addChild(aAspectLineTrans);
	cAttr.addChild(aAspectDisplaySafeAction);
	cAttr.addChild(aAspectDisplaySafeTitle);
	cAttr.setArray(true);
	cAttr.setIndexMatters(true);

	aTextType = eAttr.create("textType", "ttyp", 0);
	eAttr.addField("String", 0);
	eAttr.addField("Lens", 1);
	eAttr.addField("Camera", 2);
	eAttr.addField("Frame", 3);
	eAttr.addField("Frame And Range", 4);
	eAttr.addField("Frame Start", 5);
	eAttr.addField("Frame End", 6);
	eAttr.addField("Frame Range", 7);
	eAttr.addField("User", 8);
	eAttr.addField("Current File",9);
	eAttr.addField("File Path", 10);
	eAttr.addField("File Name", 11);
	eAttr.addField("Date", 12);
	eAttr.setInternal(true);

	aTextStr = tAttr.create("textStr", "tstr", MFnStringData::kString);
	tAttr.setConnectable(false);
	tAttr.setDefault(defaultTextAttr);
	tAttr.setInternal(true);
	
	aTextSize = eAttr.create("textSize", "ts", 0);
	eAttr.addField("Small", 0);
	eAttr.addField("Medium", 1);
	eAttr.addField("Large", 2);
	eAttr.setInternal(true);

	aTextAlign = eAttr.create("textAlign", "ta", 0);
	eAttr.addField("Left", 0);
	eAttr.addField("Center", 1);
	eAttr.addField("Right", 2);
	eAttr.setInternal(true);

	aTextPosX = nAttr.create(
					"textPosX",
					"tpx",
					MFnNumericData::kFloat,
					0.0
					);
	nAttr.setInternal(true);

	aTextPosY = nAttr.create(
					"textPosY",
					"tpy",
					MFnNumericData::kFloat,
					0.0
					);
	nAttr.setInternal(true);

	aTextPos = nAttr.create("textPos", "tp", aTextPosX, aTextPosY);

	aTextPosRel = eAttr.create("textPosRel", "tpr", 0);
	eAttr.addField("Bottom Left", 0);
	eAttr.addField("Bottom Center", 1);
	eAttr.addField("Bottom Right", 2);
	eAttr.addField("Middle Left", 3);
	eAttr.addField("Center", 4);
	eAttr.addField("Middle Right", 5);
	eAttr.addField("Top Left", 6);
	eAttr.addField("Top Center", 7);
	eAttr.addField("Top Right", 8);
	eAttr.setInternal(true);

	aTextLevel = eAttr.create("textLevel", "tl", 0);
	eAttr.addField("Window", 0);
	eAttr.addField("Filmback", 1);
	eAttr.addField("Aspect Ratio", 2);
	eAttr.setInternal(true);

	aTextARLevel = nAttr.create(
					"textARLevel",
					"tarl",
					MFnNumericData::kInt,
					0
					);
	nAttr.setInternal(true);

	aTextColor = nAttr.createColor("textColor", "tc");
	nAttr.setDefault(1.0, 1.0, 1.0);
	nAttr.setUsedAsColor(true);
	nAttr.setInternal(true);

	aTextTrans = nAttr.create(
					"textTrans",
					"tt",
					MFnNumericData::kFloat,
					0.5
					);
	nAttr.setMin(0.0);
	nAttr.setMax(1.0);
	nAttr.setInternal(true);

	aText = cAttr.create("text", "txt");
	cAttr.addChild(aTextType);
	cAttr.addChild(aTextStr);
	cAttr.addChild(aTextSize);
	cAttr.addChild(aTextAlign);
	cAttr.addChild(aTextPos);
	cAttr.addChild(aTextPosRel);
	cAttr.addChild(aTextLevel);
	cAttr.addChild(aTextARLevel);
	cAttr.addChild(aTextColor);
	cAttr.addChild(aTextTrans);
	cAttr.setArray(true);
	cAttr.setIndexMatters(true);

	aLineColor = nAttr.createColor("lineColor", "lc");
	nAttr.setDefault(1.0, 1.0, 0.0);
	nAttr.setUsedAsColor(true);
	nAttr.setInternal(true);

	aLineTrans = nAttr.create(
					"lineTrans",
					"lt",
					MFnNumericData::kFloat,
					0.5
					);
	nAttr.setMin(0.0);
	nAttr.setMax(1.0);
	nAttr.setInternal(true);

	aDisplayLineH = nAttr.create(
						"displayLineH",
						"dlh",
						MFnNumericData::kBoolean,
						false
						);
	nAttr.setInternal(true);

	aDisplayLineV = nAttr.create(
						"displayLineV",
						"dlv",
						MFnNumericData::kBoolean,
						false
						);
	nAttr.setInternal(true);

	aDisplayThirdsH = nAttr.create(
						"displayThirdsH",
						"dth",
						MFnNumericData::kBoolean,
						false
						);
	nAttr.setInternal(true);

	aDisplayThirdsV = nAttr.create(
						"displayThirdsV",
						"dtv",
						MFnNumericData::kBoolean,
						false
						);
	nAttr.setInternal(true);

	aDisplayCrosshair = nAttr.create(
							"displayCrosshair",
							"dlc",
							MFnNumericData::kBoolean,
							false
							);
	nAttr.setInternal(true);

	aTime = uAttr.create("time", "tm", MFnUnitAttribute::kTime, 0.0);

	// Add Attributes
	addAttribute(aEnableTextDrawing);
	addAttribute(aDrawingEnabled);
	addAttribute(aFilmbackAperture);
	addAttribute(aAspectRatios);
	addAttribute(aText);
	addAttribute(aLineColor);
	addAttribute(aLineTrans);
	addAttribute(aDisplayLineH);
	addAttribute(aDisplayLineV);
	addAttribute(aDisplayThirdsH);
	addAttribute(aDisplayThirdsV);
	addAttribute(aDisplayCrosshair);
	addAttribute(aTime);


	return MS::kSuccess;
}

