//=============================================================================
//
// Copyright (c) 2009, Sony Pictures Imageworks
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the
// distribution.  Neither the name of Sony Pictures Imageworks nor the
// names of its contributors may be used to endorse or promote
// products derived from this software without specific prior written
// permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
//=============================================================================
#ifndef _Reticle_h_
#define _Reticle_h_

#include <maya/MPxLocatorNode.h>
#include <maya/MFnCamera.h>
#include <maya/MMatrix.h>
#include <maya/MColor.h>
#include <vector>


typedef struct Geom
{
	double  x1, x2;
	double  y1, y2;
	double  x, y;
	MColor  lineColor;
	MColor  maskColor;
	bool	isValid;
} Geom;

typedef struct Aspect_Ratio
{
	double aspectRatio;
	int	displayMode;
	bool   displaySafeAction;
	bool   displaySafeTitle;

	Geom   aspectGeom;
	Geom   safeActionGeom;
	Geom   safeTitleGeom;
}Aspect_Ratio;

typedef struct Filmback
{
	double horizontalFilmAperture;
	double verticalFilmAperture;

	Geom   filmbackGeom;
	Geom   imageGeom;
}Filmback;

typedef struct Options
{
	bool   drawingEnabled;
	bool   enableTextDrawing;
	bool   displayLineH;
	bool   displayLineV;
	bool   displayThirdsH;
	bool   displayThirdsV;
	bool   displayCrosshair;
	MColor textColor;
	MColor lineColor;
}Options;

typedef struct TextData
{
	int		textType;
	MString	textStr;
	int		textSize;
	int		textAlign;
	double	textPosX;
	double	textPosY;
	int		textPosRel;
	int		textLevel;
	int		textARLevel;
	MColor	textColor;
}TextData;

class Reticle : public MPxLocatorNode
{
public:
							Reticle();
	virtual					~Reticle();
	static  void			*creator();
	static  MStatus			initialize();
	virtual bool			setInternalValueInContext(
								const MPlug &,
								const MDataHandle &,
								MDGContext &
								);

	virtual void			draw(
								M3dView &view,
								const MDagPath &path,
								M3dView::DisplayStyle style,
								M3dView::DisplayStatus status
								);

	virtual bool			excludeAsLocator() const;
	virtual void			postConstructor();

	virtual bool			isTransparent() const;
	virtual bool			drawLast() const;
	virtual bool			isBounded() const;
	virtual MBoundingBox	boundingBox() const;

public:
	static MTypeId			id;
	static MString			name;
	static MObject			aDrawingEnabled;
	static MObject			aEnableTextDrawing;
	static MObject			aFilmbackAperture;
	static MObject			aHorizontalFilmAperture;
	static MObject			aVerticalFilmAperture;
	static MObject			aAspectRatios;
	static MObject			aAspectRatio;
	static MObject			aDisplayMode;
	static MObject			aAspectMaskColor;
	static MObject			aAspectMaskTrans;
	static MObject			aAspectLineColor;
	static MObject			aAspectLineTrans;
	static MObject			aAspectDisplaySafeAction;
	static MObject			aAspectDisplaySafeTitle;
	static MObject			aDisplayLineH;
	static MObject			aDisplayLineV;
	static MObject			aDisplayThirdsH;
	static MObject			aDisplayThirdsV;
	static MObject			aDisplayCrosshair;
	static MObject			aLineColor;
	static MObject			aLineTrans;
	static MObject			aTime;
	static MObject			aText;
	static MObject			aTextType;
	static MObject			aTextStr;
	static MObject			aTextSize;
	static MObject			aTextAlign;
	static MObject			aTextPos;
	static MObject			aTextPosX;
	static MObject			aTextPosY;
	static MObject			aTextPosRel;
	static MObject			aTextLevel;
	static MObject			aTextARLevel;
	static MObject			aTextColor;
	static MObject			aTextTrans;

private:
	MStatus		getFilmbackData();
	MStatus		getAspectRatioChildren(
					MPlug arPlug,
					Aspect_Ratio &ar
					);

	static bool aspectRatioSortPredicate(
					const Aspect_Ratio &,
					const Aspect_Ratio&
					);

	MStatus		getAspectRatioData();
	bool		needToUpdateAspectRatios();
	MStatus		getTextChildren(MPlug tPlug, TextData &td);
	MStatus		getTextData();
	bool		needToUpdateTextData();
	MStatus		getOptions();
	MPoint		getPoint(float x, float y, M3dView &view);
	MStatus		getColor(MObject colorObj, MObject transObj, MColor &color);
	MMatrix		getMatrix(MString matrixStr);
	void		calcPortGeom();
	void		calcFilmbackGeom();
	void		calcMaskGeom(
					Geom &g,
					double w,
					double h,
					Geom &gSrc,
					double wSrc,
					double hSrc
					);

	void		calcSafeActionGeom(Aspect_Ratio &ar);
	void		calcSafeTitleGeom(Aspect_Ratio &ar);
	void		calcAspectGeom(Aspect_Ratio &ar);
	void		drawMask(Geom g1, Geom g2, MColor color, bool sides);
	void		drawLine(
					double x1,
					double x2,
					double y1,
					double y2,
					MColor color,
					bool stipple
					);

	void		drawLines(Geom g, MColor color, bool sides, bool stipple);
	void		drawText(
					MString text,
					double tx, double ty,
					void *font, MColor color, int textAlign, int textValign,
					M3dView &view
					);

	void		drawCustomTextElements(M3dView &view);

	Filmback	oFilmback, filmback;
	Geom		portGeom;
	Options		options;
	double		portWidth;
	double		portHeight;
	double		overscan;
	double		ncp;
	MMatrix		wim;
	MFnCamera	camera;
	MObject		thisNode;
	int			numAspectRatios;
	bool		needRefresh;
	std::vector<Aspect_Ratio>	ars;
	std::vector<TextData>		text;
};


#endif